###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Just a module with some utilities."""
import os
import tarfile
import zipfile
import math
import time

from DIRAC import S_OK, S_ERROR, gConfig, gLogger
from DIRAC.ConfigurationSystem.Client.Helpers import Resources
from DIRAC.Core.Utilities.List import fromChar
from DIRAC.WorkloadManagementSystem.Client.CPUNormalization import getCPUTime
from DIRAC.WorkloadManagementSystem.Utilities.JobParameters import getNumberOfJobProcessors
from LHCbDIRAC.Core.Utilities.XMLTreeParser import XMLTreeParser


def tarFiles(outputFile, files=None, compression="gz", deleteInput=False):
    """just make a tar."""
    if files is None:
        files = []

    try:
        tar = tarfile.open(outputFile, "w:" + compression)
        for fileIn in files:
            tar.add(fileIn)
        tar.close()
    except tarfile.CompressionError:
        return S_ERROR("Compression not available")

    if deleteInput:
        for fileIn in files:
            try:
                os.remove(fileIn)
            except OSError:
                pass

    return S_OK()


def zipFiles(outputFile, files=None, directory=None, deleteInput=False):
    """just make a zip.

    :param str outputFile: output file name, normally something.zip
    :param list files: file names to be added to the zip file
    :param str directory: optional directory inside the zip
    :param bool deleteInput: if you want to delete the inputs
    """
    if files is None:
        files = []

    try:
        with zipfile.ZipFile(outputFile, "w") as zipped:
            for fileIn in files:
                # ZIP does not support timestamps before 1980, so for those we simply "touch"
                st = os.stat(fileIn)
                mtime = time.localtime(st.st_mtime)
                dateTime = mtime[0:6]
                if dateTime[0] < 1980:
                    os.utime(fileIn, None)  # same as "touch"

                if directory:
                    zipped.write(fileIn, directory + "/" + fileIn, zipfile.ZIP_DEFLATED)
                else:
                    zipped.write(fileIn)
    except zipfile.LargeZipFile as zlz:
        gLogger.error("Too large file", repr(zlz))
        return S_ERROR("Too large file...?")

    if deleteInput:
        for fileIn in files:
            try:
                os.remove(fileIn)
            except OSError:
                pass

    return S_OK()


def lowerExtension():
    """Lowers the file extension of the produced files (on disk!).

    E.g.: fileName.EXTens.ION -> fileName.extens.ion
    """

    filesInDir = [x for x in os.listdir(".") if not os.path.isdir(x)]

    lowers = []

    for fileIn in filesInDir:
        splitted = fileIn.split(".")
        if len(splitted) > 1:
            lowered = ""
            for toBeLowered in splitted[1:]:
                lowered = lowered + "." + toBeLowered.lower()
                final = splitted[0] + lowered
        else:
            final = splitted[0]
        lowers.append((fileIn, final))

    for fileIn in lowers:
        os.rename(fileIn[0], fileIn[1])


#############################################################################


def getEventsToProduce(CPUe, CPUTime=None, CPUNormalizationFactor=None, maxNumberOfEvents=None, jobMaxCPUTime=None):
    """Returns the number of events to produce considering the CPU time
    available. CPUTime and CPUNormalizationFactor are taken from the LocalSite
    configuration if not provided. No checks are made on the values passed !

    Limits can be set.
    """

    if CPUNormalizationFactor is None:
        CPUNormalizationFactor = gConfig.getValue("/LocalSite/CPUNormalizationFactor", 1.0)

    if CPUTime is None:
        CPUTime = getCPUTime(CPUNormalizationFactor)
    if jobMaxCPUTime:
        gLogger.verbose(
            "CPUTimeLeft from the WN perspective: %d; Job maximum CPUTime (in seconds): %d" % (CPUTime, jobMaxCPUTime)
        )
        CPUTime = min(CPUTime, jobMaxCPUTime)

    gLogger.verbose("CPUTime = %d, CPUNormalizationFactor = %f, CPUe = %d" % (CPUTime, CPUNormalizationFactor, CPUe))

    eventsToProduce = int(math.floor(CPUTime * CPUNormalizationFactor) / float(CPUe))
    gLogger.verbose("Without limits, we can produce %d events" % eventsToProduce)

    gLogger.info("We can produce %d events" % eventsToProduce)
    willProduce = int(eventsToProduce * 0.75)
    gLogger.info("But we take a conservative approach, so 75%% of those: %d" % willProduce)

    if maxNumberOfEvents:
        gLogger.verbose("Limit for MaxNumberOfEvents: %d" % maxNumberOfEvents)
        willProduce = min(willProduce, maxNumberOfEvents)

    if willProduce < 1:
        raise RuntimeError("No time left to produce events")

    return willProduce


#############################################################################


def getCPUNormalizationFactorAvg():
    """Returns the average HS06 CPU normalization factor for the LCG sites (all
    CEs, all queues).

    Raises an Exception if it can not.
    """

    factorsSum = 0.0
    nQueues = 0

    sites = gConfig.getSections("Resources/Sites/LCG")
    if not sites["OK"]:
        raise RuntimeError(sites["Message"])
    else:
        sites = sites["Value"]

    queuesRequest = Resources.getQueues(sites)
    if not queuesRequest["OK"]:
        raise RuntimeError(queuesRequest["Message"])
    else:
        queuesRequest = queuesRequest["Value"]

    for site in queuesRequest.values():
        for ce in site.values():
            if "Queues" in ce:
                for queue in ce["Queues"].values():
                    if "SI00" in queue:
                        # convert from SI00 to HS06
                        factorsSum += float(queue["SI00"]) / 250
                        nQueues += 1

    if nQueues == 0:
        raise RuntimeError("No queues to get CPU normalization factor from")
    else:
        CPUNormalizationFactorAvg = float(factorsSum / nQueues)

    return CPUNormalizationFactorAvg


###############################################################################


def getProductionParameterValue(productionXML, parameterName):
    """Get a parameter value from a production XML description."""

    # lets assume no parameters are different only by case, as it would be a bad idea
    parameterName = parameterName.lower()

    parser = XMLTreeParser()
    parser.parseString(productionXML)
    tree = parser.tree.pop()

    for parameterElement in tree.childrens("Parameter"):
        if parameterElement.attributes["name"].lower() == parameterName:
            valueElement = parameterElement.children
            if not valueElement:
                return None
            valueElement = valueElement[0]

            return valueElement.value

    return None


###############################################################################


def getNumberOfProcessorsToUse(jobID, payloadProcessors=None):
    """get the number of processors to use for an application step on a certain node.

    payloadProcessors corresponds normally to the workflow (and JDL) parameter "MaxNumberOfProcessors"
    """

    siteName = gConfig.getValue("/LocalSite/Site")
    gridCE = gConfig.getValue("/LocalSite/GridCE")
    queue = gConfig.getValue("/LocalSite/CEQueue")

    if _multicoreWN(siteName, gridCE, queue):
        maxNumberOfProcessorsAllowedOnTheWN = int(getNumberOfJobProcessors(jobID))
        if not payloadProcessors:
            payloadProcessors = maxNumberOfProcessorsAllowedOnTheWN

        return min(maxNumberOfProcessorsAllowedOnTheWN, int(payloadProcessors))
        # NB: the case with
        # maxNumberOfProcessorsAllowedOnTheWN < number of processors requested (JDL param "NumberOfProcessors")
        # should not happen by construction, as the job should not be matched in a first place

    else:
        gLogger.info("Running in MultiProcessor is not allowed here")
        return 1


def _multicoreWN(siteName, gridCE, queue):
    """Returns "True" if the CE, or the Queue is marked as one where multi-
    processing is allowed (by having Tag "MultiProcessor")"""
    # Tags of the CE
    tags = fromChar(gConfig.getValue(f"/Resources/Sites/{siteName.split('.')[0]}/{siteName}/CEs/{gridCE}/Tag", ""))
    # Tags of the Queue
    tags += fromChar(
        gConfig.getValue(f"/Resources/Sites/{siteName.split('.')[0]}/{siteName}/CEs/{gridCE}/Queues/{queue}/Tag", "")
    )

    return bool(tags and "MultiProcessor" in tags)



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Retrieve from Bookkeeping information for a given run."""
import DIRAC
from DIRAC import gLogger, S_OK
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.registerSwitch("", "Production=", "   <prodID>, get the run list from a production")
    Script.registerSwitch("", "Active", "   only get Active runs")
    Script.registerSwitch("", "Information=", "   <item> returns only the relevant information item")
    Script.registerSwitch(
        "", "ByValue", "   if set, the information is a list of runs for each value of the information item"
    )
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            ["Usage:", f"  {Script.scriptName} [option|cfgfile] ... Run", "Arguments:", "  Run:      Run Number"]
        )
    )
    Script.parseCommandLine(ignoreErrors=True)
    runRanges = []
    for arg in Script.getPositionalArgs():
        runRanges += arg.split(",")

    runSet = set()
    for run in runRanges:
        try:
            if ":" in arg:
                run1, run2 = run.split(":")
                runSet.update(range(int(run1), int(run2) + 1))
            else:
                runSet.add(int(run))
        except (ValueError, IndexError) as e:
            gLogger.exception("Invalid run number", arg, lException=e)
            DIRAC.exit(1)

    production = None
    item = None
    byValue = False
    active = False
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "Production":
            try:
                production = [int(prod) for prod in switch[1].split(",")]
            except ValueError as e:
                gLogger.exception("Bad production ID", lException=e)
                DIRAC.exit(1)
        elif switch[0] == "Information":
            item = switch[1]
        elif switch[0] == "ByValue":
            byValue = True
        elif switch[0] == "Active":
            active = True

    from LHCbDIRAC.DataManagementSystem.Client.DMScript import printDMResult, ProgressBar
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    bk = BookkeepingClient()

    if production:
        from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

        trClient = TransformationClient()
        condDict = {"TransformationID": production}
        if active:
            condDict["Status"] = "Active"
        res = trClient.getTransformationRuns(condDict)
        if res["OK"]:
            runSet.update(run["RunNumber"] for run in res["Value"])
        else:
            gLogger.fatal("Error getting production runs", res["Message"])
            DIRAC.exit(2)
        gLogger.notice(f"Found {len(runSet)} runs")

    # Use this call to get information but also the actual list of existing runs
    res = bk.getRunStatus(list(runSet))
    if not res["OK"]:
        gLogger.fatal("Error getting the run info", res["Message"])
        DIRAC.exit(2)
    runStatus = res["Value"]["Successful"]

    sep = ""
    if item:
        result = {"Successful": {}, "Failed": {}}
        success = result["Successful"]
        failed = result["Failed"]
        if item == "Finished":
            # We can get it in a single call
            for run in runStatus:
                finished = {"N": "No", "Y": "Yes"}.get(runStatus.get(run, {}).get("Finished"), "Unknown")
                if byValue:
                    success.setdefault(finished, []).append(run)
                else:
                    success[run] = finished
        else:
            progressBar = ProgressBar(len(runStatus), title=f"Getting {item} for {len(runStatus)} runs", step=10)

    if item != "Finished":
        for run in sorted(runStatus):
            if item:
                progressBar.loop()
            finished = {"N": "No", "Y": "Yes"}.get(runStatus[run].get("Finished"), "Unknown")

            res = bk.getRunInformations(run)
            if res["OK"]:
                info = res["Value"]
                if item:
                    if item not in info:
                        gLogger.error("Item not found", f"\n Valid items: {str(sorted(info))}")
                        DIRAC.exit(3)
                    itemValue = info.get(item, "Unknown")
                    if byValue:
                        success.setdefault(itemValue, []).append(run)
                    else:
                        result["Successful"][run] = itemValue
                    continue
                runstart = info.get("RunStart", "Unknown")
                runend = info.get("RunEnd", "Unknown")
                configname = info.get("Configuration Name", "Unknown")
                configversion = info.get("Configuration Version", "Unknown")
                fillnb = info.get("FillNumber", "Unknown")
                datataking = info.get("DataTakingDescription", "Unknown")
                processing = info.get("ProcessingPass", "Unknown")
                stream = info.get("Stream", "Unknown")
                fullstat = info.get("FullStat", "Unknown")
                nbofe = info.get("Number of events", "Unknown")
                nboff = info.get("Number of file", "Unknown")
                fsize = info.get("File size", "Unknown")
                totalLuminosity = info.get("TotalLuminosity", "Unknown")
                tck = info.get("Tck", "Unknown")

                if sep:
                    print(sep)
                print("Run  Informations for run %d: " % run)
                print("Run Start:".ljust(30), str(runstart))
                print("Run End:".ljust(30), str(runend))
                print("Total luminosity:".ljust(30), str(totalLuminosity))
                print("  Configuration Name:".ljust(30), configname)
                print("  Configuration Version:".ljust(30), configversion)
                print("  FillNumber:".ljust(30), fillnb)
                print("  Finished:".ljust(30), finished)
                print("  Data taking description:".ljust(30), datataking)
                print("  Processing pass:".ljust(30), processing)
                print("  TCK:".ljust(30), tck)
                print("  Stream:".ljust(30), stream)
                just = len(str(fsize)) + 3
                print("  FullStat:".ljust(30), str(fullstat).ljust(just), " Total: ".ljust(10) + str(sum(fullstat)))
                print("  Number of events:".ljust(30), str(nbofe).ljust(just), " Total:".ljust(10) + str(sum(nbofe)))
                print("  Number of files:".ljust(30), str(nboff).ljust(just), " Total: ".ljust(10) + str(sum(nboff)))
                print("  File size:".ljust(30), str(fsize).ljust(just), " Total: ".ljust(10) + str(sum(fsize)))
                sep = 20 * "="
            elif item:
                failed[run] = res["Message"]
        if item:
            progressBar.endLoop()

    if item:
        if not failed:
            del failed
        if byValue:
            for val in success:
                success[val] = f"({len(success[val])} runs) - " + (",".join(sorted(str(run) for run in success[val])))
        printDMResult(S_OK(result), empty="None", script="dirac-bookkeeping-run-information")


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Database class for storing information about Analysis Productions

For more information on the meaning of the various objects see :py:mod:`.AnalysisProductionsClient`.

Tables are defined using SQLAlchemy and imported from :py:mod:`.AnalysisProductionsObjects`.
Example usage of this class can be found in `Test_AnalysisProductionsDB.py`.
"""
import functools
import datetime
from collections import defaultdict
from copy import deepcopy
from contextlib import contextmanager

from sqlalchemy import create_engine, delete, func, insert, or_, tuple_, select, case, JSON, cast, Integer, and_
from sqlalchemy.orm.attributes import flag_modified
from sqlalchemy.orm import Session

from DIRAC.ConfigurationSystem.Client.Utilities import getDBParameters
from DIRAC.Core.Base.DIRACDB import DIRACDB
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise

from LHCbDIRAC.ProductionManagementSystem.DB.AnalysisProductionsObjects import (
    Base,
    AnalysisSample as AP,
    AutoTag,
    Tag,
    Request,
    AnalysisOwner,
    Publication,
)
from LHCbDIRAC.ProductionManagementSystem.DB.extra_func import json_group_array, json_group_object


def inject_session(func):
    """Decorator to inject the session into a class method

    Decorator to start a SQLAlchemy Session and inject it in the wrapped function
    as a keyword argument.
    """

    @functools.wraps(func)
    def new_func(self, *args, **kwargs):
        if "session" in kwargs:
            raise NotImplementedError("session cannot be passed through the inject_session decorator")
        with self.session as session:
            return func(self, *args, **kwargs, session=session)

    return new_func


class AnalysisProductionsDB(DIRACDB):
    __engineCache = {}

    def __init__(self, *, url=None, parentLogger=None):
        self.fullname = self.__class__.__name__
        super().__init__(parentLogger=parentLogger)
        if url is None:
            param = returnValueOrRaise(getDBParameters("ProductionManagement/AnalysisProductionsDB"))
            url = f"mysql://{param['User']}:{param['Password']}@{param['Host']}:{param['Port']}/{param['DBName']}"
        self.setURL(url)

    def setURL(self, url):
        if url not in self.__engineCache or ":memory:" in url:
            engine = create_engine(
                url, pool_recycle=3600, echo_pool=True, echo=self.log.getLevel() == "DEBUG", future=True
            )
            Base.metadata.create_all(engine)
            self.__engineCache[url] = engine
        self.engine = self.__engineCache[url]

    @property
    @contextmanager
    def session(self):
        with Session(self.engine, future=True) as session, session.begin():
            yield session

    @inject_session
    def listAnalyses(self, *, at_time=None, session: Session):
        query = _filterForTime(session.query(AP.wg, AP.analysis).distinct(), AP, at_time)
        result = defaultdict(list)
        for wg, analysis in query:
            result[wg].append(analysis)
        return dict(result)

    @inject_session
    def listAnalyses2(self, *, at_time=None, session: Session):
        # Select the owners of the analyses
        query = select(AnalysisOwner.wg, AnalysisOwner.analysis, AnalysisOwner.username).distinct()
        owners = defaultdict(lambda: defaultdict(list))
        for wg, analysis, username in session.execute(query).all():
            owners[wg][analysis].append(username)

        # Select the analyses and build the list of results
        query = select(
            AP.wg,
            AP.analysis,
            # Add a column n_{STATE_NAME} containing the number of analyses in each possible state
            *(
                # MySQL returns a decimal type from sum() so we need to cast it to an integer
                cast(func.sum(case((AP.state == state, 1), else_=0)), Integer).label(  # pylint: disable=not-callable
                    f"n_{state}"
                )
                for state in AP.VALID_STATES
            ),
            func.count().label("n_total"),  # pylint: disable=not-callable
            func.min(AP.housekeeping_interaction_due).label("earliest_housekeeping_due"),
        )
        query = query.group_by(AP.wg, AP.analysis)
        query = _filterForTime(query, AP, at_time)
        return [dict(row._mapping) | {"owners": owners[row.wg][row.analysis]} for row in session.execute(query).all()]

    @inject_session
    def listRequests(self, *, session: Session):
        analyses = select(
            AP.request_id,
            AP.filetype,
            AP.name,
            AP.version,
            json_group_array(func.json_array(AP.wg, AP.analysis)).label("analyses"),
        )
        analyses = analyses.group_by(AP.request_id, AP.filetype, AP.name, AP.version)
        analyses = _filterForTime(analyses, AP, at_time=None)
        analyses = analyses.subquery(name="requests")

        autotags = select(
            Request.request_id,
            Request.filetype,
            # Can't use AGG_FUNC.filter as it's not supported by mysql
            case(
                (
                    func.sum(AutoTag.name.is_not(None)) == 0,  # pylint: disable=not-callable
                    func.json_object(type_=JSON),  # pylint: disable=not-callable
                ),
                else_=json_group_object(AutoTag.name, AutoTag.value),
            ).label("autotags"),
        )
        autotags = autotags.join(
            AutoTag,
            tuple_(Request.request_id, Request.filetype) == tuple_(AutoTag.request_id, AutoTag.filetype),
            isouter=True,
        )
        autotags = autotags.group_by(Request.request_id, Request.filetype)
        autotags = autotags.subquery(name="autotags")

        tags = select(
            AP.request_id,
            AP.filetype,
            # Can't use AGG_FUNC.filter as it's not supported by mysql
            case(
                (
                    func.sum(Tag.name.is_not(None)) == 0,  # pylint: disable=not-callable
                    func.json_array(type_=JSON),  # pylint: disable=not-callable
                ),
                else_=json_group_array(func.json_array(Tag.name, Tag.value)),
            ).label("tags"),
        ).distinct()
        tags = tags.join(Tag, AP.sample_id == Tag.sample_id, isouter=True)
        tags = tags.group_by(AP.request_id, AP.filetype)
        tags = tags.subquery(name="tags")

        query = select(analyses, autotags.c.autotags, tags.c.tags)
        query = query.join(
            autotags,
            tuple_(analyses.c.request_id, analyses.c.filetype) == tuple_(autotags.c.request_id, autotags.c.filetype),
            isouter=True,
        )
        query = query.join(
            tags,
            tuple_(analyses.c.request_id, analyses.c.filetype) == tuple_(tags.c.request_id, tags.c.filetype),
            isouter=True,
        )
        return [dict(row._mapping) for row in session.execute(query).all()]

    @inject_session
    def getOwners(self, *, wg=None, analysis=None, session: Session):
        query = select(AnalysisOwner.username)
        query = query.where(AnalysisOwner.wg == wg)
        query = query.where(AnalysisOwner.analysis == analysis)
        return session.execute(query).scalars().all()

    @inject_session
    def setOwners(self, *, wg=None, analysis=None, owners=None, session: Session):
        query = select(AnalysisOwner.username)
        query = query.where(AnalysisOwner.wg == wg)
        query = query.where(AnalysisOwner.analysis == analysis)
        existing_owners = session.execute(query).scalars().all()

        new_owners = [
            {"wg": wg, "analysis": analysis, "username": owner} for owner in owners if owner not in existing_owners
        ]
        if new_owners:
            session.execute(insert(AnalysisOwner).values(new_owners))

        query = delete(AnalysisOwner)
        query = query.where(AnalysisOwner.wg == wg)
        query = query.where(AnalysisOwner.analysis == analysis)
        query = query.where(~AnalysisOwner.username.in_(owners))
        session.execute(query)

    @inject_session
    def getProductions(
        self,
        *,
        wg=None,
        analysis=None,
        version=None,
        name=None,
        state=None,
        at_time=None,
        show_archived=False,
        require_has_publication=False,
        session: Session,
    ):
        query = select(
            AP.wg,
            AP.analysis,
            AP.sample_id,
            AP.validity_start,
            AP.validity_end,
            AP.name,
            AP.version,
            AP.request_id,
            AP.filetype,
            AP.state,
            AP.last_state_update,
            AP.extra_info["transformations"].label("transformations"),
            AP.progress,
            AP.extra_info["jira_task"].label("jira_task"),
            AP.extra_info["merge_request"].label("merge_request"),
            AP.housekeeping_interaction_due,
        )
        query = query.filter(*_buildCondition(wg, analysis, name, version))
        if state is not None:
            query = query.filter(AP.state == state)
        if not show_archived:
            query = _filterForTime(query, AP, at_time)

        query = query.group_by(AP.sample_id).subquery(name="samples")

        pub_q = select(
            Publication.sample_id,
            case(
                (
                    func.count(Publication.number) == 0,  # pylint: disable=not-callable
                    func.json_array(type_=JSON),  # pylint: disable=not-callable
                ),
                else_=json_group_array(Publication.number),
            ).label("publications"),
        )
        pub_q = pub_q.group_by(Publication.sample_id).subquery(name="pubs")

        mque = select(query, pub_q.c.publications)
        mque = mque.join(pub_q, pub_q.c.sample_id == query.c.sample_id, isouter=True)

        if require_has_publication:
            # Return a sample only if it has a publication number assigned to it
            mque = mque.filter(func.json_array_length(pub_q.c.publications) > 0)

        results = []
        for row in session.execute(mque).all():
            result = {
                "name": row.name,
                "version": row.version,
                "request_id": row.request_id,
                "filetype": row.filetype,
                "state": row.state,
                "last_state_update": row.last_state_update,
                "transformations": row.transformations,
            }
            if row.progress is not None:
                result["progress"] = row.progress
            if row.jira_task is not None:
                result["jira_task"] = row.jira_task
            if row.merge_request is not None:
                result["merge_request"] = row.merge_request
            result.update(
                {
                    "wg": row.wg,
                    "analysis": row.analysis,
                    "sample_id": row.sample_id,
                    # TODO: Remove
                    "owners": [],
                    "validity_start": row.validity_start,
                    "validity_end": row.validity_end,
                    "housekeeping_interaction_due": row.housekeeping_interaction_due,
                    "publications": row.publications or [],
                }
            )
            results.append(result)
        return results

    @inject_session
    def getArchivedRequests(self, *, state=None, session: Session):
        sq = (
            session.query(AP.request_id, AP.filetype)
            .filter(or_(AP.validity_end.is_(None), datetime.datetime.now() < AP.validity_end))
            .distinct()
            .subquery()
        )
        query = session.query(Request).filter(~tuple_(Request.request_id, Request.filetype).in_(select(sq)))
        if state is not None:
            query = query.filter(AP.state == state)
        return [result.toDict() for result in query]

    @inject_session
    def getTags(self, wg, analysis, *, at_time=None, session: Session):
        return _getTags(session, wg=wg, analysis=analysis, at_time=at_time)

    @inject_session
    def getKnownAutoTags(self, *, session) -> set:
        return _getKnownAutoTags(session)

    @inject_session
    def registerTransformations(self, transforms: dict[int, dict[str, list[dict]]], *, session: Session):
        if not transforms:
            raise ValueError("No transforms passed")
        transforms = deepcopy({(prid, ft): ts for prid, x in transforms.items() for ft, ts in x.items()})
        for request in session.query(Request).filter(tuple_(Request.request_id, Request.filetype).in_(transforms)):
            knownTransforms = {t["id"] for t in request.extra_info["transformations"]}
            for transform in transforms.pop((request.request_id, request.filetype)):
                if transform["id"] in knownTransforms:
                    raise ValueError(f"Transformation is already known {transform['id']}")
                # TODO: Validate the transform object
                request.extra_info["transformations"].append(transform)
                # By default SQLAlchemy doesn't detect changes in JSON columns when using the ORM
                # Ideally this should be fixed in the database definition but flagging manually is
                # good enough for now
                flag_modified(request, "extra_info")
        if transforms:
            raise ValueError(f"Did not find requests for IDs: {list(transforms)}")

    @inject_session
    def deregisterTransformations(self, tIDs: dict[int, dict[str, list[int]]], *, session: Session):
        """See :meth:`~.AnalysisProductionsClient.registerTransformations`"""
        if not tIDs:
            raise ValueError("No transform IDs passed")
        tIDs = deepcopy({(prid, ft): ts for prid, x in tIDs.items() for ft, ts in x.items()})
        query = session.query(Request).filter(tuple_(Request.request_id, Request.filetype).in_(tIDs))
        for request in query:
            for tID in tIDs.pop((request.request_id, request.filetype)):
                for i, transform in enumerate(request.extra_info["transformations"]):
                    if transform["id"] == tID:
                        request.extra_info["transformations"].pop(i)
                        break
                else:
                    raise ValueError(f"Transformation {tID} is not known")
                flag_modified(request, "extra_info")
        if tIDs:
            raise ValueError(f"Did not find requests for IDs: {list(tIDs)}")

    def registerRequests(self, requests: list[dict]):
        request_ids = {(r["request_id"], r["filetype"]) for r in requests}
        with self.session as session:
            known_ids = {
                (i, ft)
                for i, ft in session.query(AP.request_id, AP.filetype).filter(
                    tuple_(AP.request_id, AP.filetype).in_(request_ids)
                )
            }
            if known_ids:
                raise ValueError(f"Already registered requests: {known_ids!r}")

            for r in requests:
                self.log.info(
                    "Registering Analysis Production request",
                    f"{r['wg']} {r['analysis']} {r['version']} {r['request_id']} {r['name']}",
                )
                sample = AP(
                    request_id=r["request_id"],
                    filetype=r["filetype"],
                    name=r["name"],
                    version=r["version"],
                    wg=r["wg"],
                    analysis=r["analysis"],
                    validity_start=r["validity_start"],
                    extra_info=r["extra_info"],
                    auto_tags=[AutoTag(name=x["name"], value=x["value"]) for x in r["auto_tags"]],
                )
                # TODO: Remove
                # sample.owners = [User(username=x["username"]) for x in r["owners"]]
                session.add(sample)

        with self.session as session:
            query = session.query(AP)
            query = query.filter(tuple_(AP.request_id, AP.filetype).in_(request_ids))
            return [result.toDict() for result in query]

    @inject_session
    def addRequestsToAnalysis(self, wg: str, analysis: str, requests: list[tuple[int, str]], *, session: Session):
        self.log.info("Adding samples to analysis", f"({wg}/{analysis}) {','.join(map(str, requests))}")

        query = select(AP.request_id, AP.filetype, AP.sample_id).filter(
            AP.wg == wg,
            AP.analysis == analysis,
            tuple_(AP.request_id, AP.filetype).in_(requests),
            AP.validity_end.is_(None),
        )
        if already_existing := session.execute(query).all():
            raise ValueError(
                f"Some requests are already registered for {wg}/{analysis} request_id->sample_id mapping is "
                f"{ {(request_id, filetype): sample_id for request_id, filetype, sample_id in already_existing} }"
            )

        query = insert(AP).values(
            [{"wg": wg, "analysis": analysis, "request_id": request_id, "filetype": ft} for request_id, ft in requests]
        )
        session.execute(query)

    @inject_session
    def archiveSamples(self, sample_ids: list[int], *, session: Session):
        self.log.info("Archiving Analysis Productions", ",".join(map(str, sample_ids)))
        query = session.query(AP.sample_id)
        query = query.filter(AP.sample_id.in_(sample_ids))
        known_sample_ids = {i for i, in query}
        if len(known_sample_ids) != len(sample_ids):
            raise ValueError(f"Unknown sample IDs passed {known_sample_ids - set(sample_ids)!r}")
        query = query.filter(AP.validity_end.is_(None))
        known_sample_ids = {i for i, in query}
        if len(known_sample_ids) != len(sample_ids):
            raise ValueError(f"Some samples have already been archived {known_sample_ids - set(sample_ids)!r}")
        query = session.query(AP).filter(AP.sample_id.in_(sample_ids))
        query.update({"validity_end": func.now()})  # pylint: disable=not-callable

    @inject_session
    def archiveSamplesAtSpecificTime(self, sample_ids: list[int], archive_time: datetime.datetime, *, session: Session):
        self.log.info(f"Archiving Analysis Productions at {archive_time}", ",".join(map(str, sample_ids)))
        query = session.query(AP.sample_id)
        query = query.filter(AP.sample_id.in_(sample_ids))
        known_sample_ids = {i for i, in query}
        if len(known_sample_ids) != len(sample_ids):
            raise ValueError(f"Unknown sample IDs passed {known_sample_ids - set(sample_ids)!r}")
        query = query.filter(AP.validity_end.is_(None))
        known_sample_ids = {i for i, in query}
        if len(known_sample_ids) != len(sample_ids):
            raise ValueError(f"Some samples have already been archived {known_sample_ids - set(sample_ids)!r}")
        query = session.query(AP).filter(AP.sample_id.in_(sample_ids))

        query.update({"validity_end": archive_time})  # pylint: disable=not-callable

    @inject_session
    def delayHousekeepingInteractionDue(
        self, sample_ids: list[int], next_interaction_due: datetime.datetime, *, session: Session
    ):
        self.log.info(
            "Delaying next Analysis Productions housekeeping interaction",
            f"to {next_interaction_due} for Analysis Production IDs {','.join(map(str, sample_ids))}",
        )
        query = session.query(AP.sample_id)
        query = query.filter(AP.sample_id.in_(sample_ids))

        known_sample_ids = {i for i, in query}
        if len(known_sample_ids) != len(sample_ids):
            raise ValueError(f"Unknown sample IDs passed {known_sample_ids - set(sample_ids)!r}")
        query = query.filter(AP.validity_end.is_(None))  # don't bother with archived samples
        known_sample_ids = {i for i, in query}
        if len(known_sample_ids) != len(sample_ids):
            raise ValueError(f"Some samples have already been archived {known_sample_ids - set(sample_ids)!r}")
        query = session.query(AP).filter(AP.sample_id.in_(sample_ids))

        query.update({"housekeeping_interaction_due": next_interaction_due})  # pylint: disable=not-callable

    @inject_session
    def getHousekeepingInteractionDueNow(self, *, session: Session):
        query = select(
            AP.wg,
            AP.analysis,
            case(
                (
                    func.sum(AP.housekeeping_interaction_due.is_not(None)) == 0,  # pylint: disable=not-callable
                    func.json_object(type_=JSON),  # pylint: disable=not-callable
                ),
                else_=json_group_object(AP.sample_id, AP.housekeeping_interaction_due),
            ).label("samples_due"),
        )
        # confirm it is not archived
        query = query.filter(AP.validity_end.is_(None))
        # confirm due date is in the past
        query = query.filter(AP.housekeeping_interaction_due <= func.now())  # pylint: disable=not-callable
        query = query.group_by(AP.wg, AP.analysis)
        query = query.subquery(name="samples")

        owner = select(
            AnalysisOwner.wg,
            AnalysisOwner.analysis,
            json_group_array(AnalysisOwner.username).label("owner_usernames"),
        )
        owner = owner.group_by(AnalysisOwner.wg, AnalysisOwner.analysis)
        owner = owner.subquery(name="owners")

        fq = select(query, owner.c.owner_usernames)
        fq = fq.join(owner, and_(query.c.wg == owner.c.wg, query.c.analysis == query.c.analysis), isouter=True)
        samples_due = [dict(row._mapping) for row in session.execute(fq).all()]
        return samples_due

    @inject_session
    def addPublication(self, sample_ids: list[int], publication_number: str, *, session: Session):
        if len(publication_number) > 64:
            raise ValueError("This publication number is too long (>64 chars)")

        query = insert(Publication).values([{"number": publication_number, "sample_id": sid} for sid in sample_ids])
        session.execute(query)

    @inject_session
    def getPublications(self, sample_ids: list[int] | None = None, *, session: Session):
        numbers = defaultdict(list)

        ap_q = select(
            AP.wg,
            AP.analysis,
            AP.sample_id,
            AP.validity_start,
            AP.validity_end,
            AP.name,
            AP.version,
            AP.request_id,
            AP.filetype,
            AP.state,
        ).subquery(name="samples")

        query = select(Publication.number, Publication.sample_id, ap_q)
        query = query.join(AP, AP.sample_id == Publication.sample_id)

        if sample_ids:
            query = query.filter(Publication.sample_id.in_(sample_ids))
        for row in session.execute(query).all():
            numbers[row.number].append(
                {
                    "sample_id": row.sample_id,
                    "request_id": row.request_id,
                    "filetype": row.filetype,
                    "wg": row.wg,
                    "analysis": row.analysis,
                    "name": row.name,
                    "version": row.version,
                    "state": row.state,
                    "validity_start": row.validity_start,
                    "validity_end": row.validity_end,
                }
            )
        return numbers

    @inject_session
    def setState(self, newState: dict[tuple[int, str], dict], *, session: Session):
        for request_id_ft, updateDict in newState.items():
            query = session.query(Request).filter(tuple_(Request.request_id, Request.filetype) == request_id_ft)
            rowsUpdated = query.update({getattr(Request, k): v for k, v in updateDict.items()})
            if rowsUpdated != 1:
                raise ValueError(
                    f"Failed to update Request({request_id_ft}) with {updateDict!r}, {rowsUpdated} matching rows found"
                )

    @inject_session
    def setTags(self, oldTags: dict[int, dict[str, str]], newTags: dict[int, dict[str, str]], *, session: Session):
        if set(oldTags) != set(newTags):
            raise ValueError("oldTags and newTags must contain the same keys")
        # Tags should always be lowercase in the database
        oldTags = {int(i): {str(k).lower(): str(v).lower() for k, v in x.items()} for i, x in oldTags.items()}
        newLengths = {int(i): len(x) for i, x in newTags.items()}
        newTags = {int(i): {str(k).lower(): str(v).lower() for k, v in x.items()} for i, x in newTags.items()}
        if newLengths != {i: len(x) for i, x in newTags.items()}:
            raise ValueError("newTags contains duplicate keys when converted to lowercase")

        # Compute what needs to be changed, while also ensuring the auto tags aren't touched
        knownAutoTags = _getKnownAutoTags(session)
        toRemove = []
        toAdd = []
        for sample_id, old in oldTags.items():
            new = newTags[sample_id]
            removed_tags = set(old) - set(new)
            modified_tags = {k: new[k] for k in set(new) & set(old) if new[k] != old[k]}
            added_tags = {k: new[k] for k in set(new) - set(old)}
            # Ensure that the automatic tags aren't being modifed
            if modifiedAutoTags := {*removed_tags, *modified_tags, *added_tags} & knownAutoTags:
                raise ValueError(f"Cannot modify AutoTags {modifiedAutoTags}")
            # Tags are modified by being removed and re-added allow for time-travel
            toRemove += [(sample_id, k) for k in {*removed_tags, *modified_tags}]
            toAdd += [(sample_id, k, v) for k, v in {**added_tags, **modified_tags}.items()]

        latestOldTags = _getTags(session, sample_ids=oldTags)
        if oldTags != latestOldTags:
            raise ValueError("oldTags is out of date")

        # Remove the old tags
        query = _filterForTime(session.query(Tag), Tag, at_time=None)
        query = query.filter(tuple_(Tag.sample_id, Tag.name).in_(toRemove))
        query.update({"validity_end": func.now()})  # pylint: disable=not-callable
        # Add the new tag values
        for sample_id, name, value in toAdd:
            session.add(Tag(sample_id=sample_id, name=name, value=value))


def _getKnownAutoTags(session: Session):
    return {name for name, in session.query(AutoTag.name).distinct()}


def _getTags(session, *, wg=None, analysis=None, at_time=None, sample_ids=None):
    results = defaultdict(dict)
    # Get the automatic tags
    query = session.query(AP.sample_id, AutoTag.name, AutoTag.value)
    query = query.filter(AP.request_id == AutoTag.request_id)
    query = query.filter(AP.filetype == AutoTag.filetype)
    if sample_ids:
        query = query.filter(AP.sample_id.in_(sample_ids))
    query = query.filter(*_buildCondition(wg, analysis))
    query = _filterForTime(query, AP, at_time)
    for sampleID, name, value in query:
        results[sampleID][name] = value
    # Get the manual tags
    query = session.query(AP.sample_id, Tag.name, Tag.value)
    query = query.filter(AP.sample_id == Tag.sample_id)
    if sample_ids:
        query = query.filter(AP.sample_id.in_(sample_ids))
    if wg is not None:
        query = query.filter(*_buildCondition(wg, analysis))
    query = _filterForTime(query, AP, at_time)
    query = _filterForTime(query, Tag, at_time)
    for sampleID, name, value in query:
        results[sampleID][name] = value
    return dict(results)


def _filterForTime(query, obj, at_time):
    if at_time is None:
        at_time = func.now()  # pylint: disable=not-callable
    return query.filter(
        obj.validity_start <= at_time,
        or_(obj.validity_end.is_(None), at_time < obj.validity_end),
    )


def _buildCondition(wg, analysis=None, name=None, version=None):
    """Build a SQLAlchemy query for the AnalysisProductions table"""
    if wg is not None:
        yield func.lower(AP.wg) == func.lower(wg)
    if analysis is not None:
        yield func.lower(AP.analysis) == func.lower(analysis)
    if name is not None:
        yield func.lower(AP.name) == func.lower(name)
    if version is not None:
        yield AP.version == version



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Uses the DM script switches, and, unless a list of LFNs is provided

1. If --Directory is used: get files in FC directories, check if they are in BK and if the replica flag is set
2. If --Production is used get files in the FC directories used, and proceed as with --Directory

If --FixFC is set, remove from SE and FC
If --FixBK is set and no replica flag, set replica flag in the BK
"""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    # Script initialization
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript
    import DIRAC

    Script.setUsageMessage(
        "\n".join(
            [
                __doc__,
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] [values]",
            ]
        )
    )
    dmScript = DMScript()
    dmScript.registerNamespaceSwitches()  # Directory
    dmScript.registerFileSwitches()  # File, LFNs
    dmScript.registerBKSwitches()
    Script.registerSwitch("", "FixBK", "   Take action to fix the BK")
    Script.registerSwitch("", "FixFC", "   Take action to fix the FC")
    Script.registerSwitch("", "NoFC2SE", "   Don't execute an FC2SE check")
    Script.registerSwitch("", "AffectedRuns", "   List the runs affected by the encountered problem")
    Script.parseCommandLine(ignoreErrors=True)

    from DIRAC import gLogger

    fixBK = False
    fixFC = False
    checkFC2SE = True
    listAffectedRuns = False
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "FixFC":
            fixFC = True
        if switch[0] == "FixBK":
            fixBK = True
        elif switch[0] == "AffectedRuns":
            listAffectedRuns = True
        elif switch[0] == "NoFC2SE":
            checkFC2SE = False

    if fixFC and fixBK:
        gLogger.notice("Can't fix both FC and BK, please choose")
        DIRAC.exit(0)
    # imports
    from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import ConsistencyChecks

    cc = ConsistencyChecks()
    cc.directories = dmScript.getOption("Directory", [])
    cc.lfns = dmScript.getOption("LFNs", []) + [lfn for arg in Script.getPositionalArgs() for lfn in arg.split(",")]
    productions = dmScript.getOption("Productions", [])
    runs = dmScript.getOption("Runs", [])

    from LHCbDIRAC.DataManagementSystem.Client.CheckExecutors import doCheckFC2BK

    if productions and not runs:
        fileType = dmScript.getOption("FileType", [])
        if fileType:
            cc.fileType = fileType
        for prod in productions:
            cc.prod = prod
            gLogger.always("Processing production %d" % cc.prod)
            doCheckFC2BK(cc, fixFC, fixBK, listAffectedRuns)
            gLogger.always("Processed production %d" % cc.prod)
    else:
        bkQuery = dmScript.getBKQuery(visible="All")
        if bkQuery:
            bkQuery.setOption("ReplicaFlag", "All")
            cc.bkQuery = bkQuery
        doCheckFC2BK(cc, fixFC=fixFC, fixBK=fixBK, listAffectedRuns=listAffectedRuns, checkFC2SE=checkFC2SE)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""stores the jobs and its parameters."""


class Job:
    """Job class."""

    def __init__(self):
        """initialize the class members."""
        self.configuration = None
        self.options = []
        self.parameters = []
        self.inputFiles = []
        self.outputFiles = []
        self.simulationCondition = None
        self.dataTakingCondition = None
        self.jobID = -1
        self.fileName = ""

    def addJobParams(self, jobParams):
        """sets the job parameters."""
        self.parameters += [jobParams]

    def exists(self, jobParam):
        """checks a given job parameter."""
        return any(i.name == jobParam for i in self.parameters)

    def getParam(self, jobParam):
        """returns a job parameter."""
        for i in self.parameters:
            if i.name == jobParam:
                return i

    def removeParam(self, jobParam):
        """removes a job parameter."""
        for i in self.parameters:
            if i.name == jobParam:
                self.parameters.remove(i)

    def addJobInputFiles(self, files):
        """adds the input file to a job."""
        self.inputFiles += [files]

    def addJobOutputFiles(self, files):
        """adds the output files to a job."""
        self.outputFiles += [files]

    def getOutputFileParam(self, paramName):
        """returns the parameters of a output file."""
        for i in self.outputFiles:
            param = i.getParam(paramName)
            if param is not None:
                return param

    def __repr__(self):
        """formats the output of the print command."""
        result = ["JOB: \n"]
        result += [str(self.configuration) + " "]
        for option in self.options:
            result += [str(option)]
        result += ["\n"]
        for param in self.parameters:
            result += [str(param)]
        result += ["\n"]
        for jobinput in self.inputFiles:
            result += [str(jobinput)]
        for output in self.outputFiles:
            result += [str(output)]
        result += ["\n"]
        return "".join(result)

    def writeToXML(self):
        """writes an XML string."""
        string = []
        string += ['<?xml version="1.0" encoding="ISO-8859-1"?>\n']
        string += ['<!DOCTYPE Job SYSTEM "book.dtd">\n']

        string += [str(self.configuration.writeToXML())]
        for param in self.parameters:
            string += [str(param.writeToXML())]

        for inputFile in self.inputFiles:
            string += [str(inputFile.writeToXML())]

        for output in self.outputFiles:
            string += [str(output.writeToXML())]

        if self.simulationCondition:
            string += [str(self.simulationCondition.writeToXML())]

        daq = self.dataTakingCondition
        if daq is not None:
            string += [str(daq.writeToXML())]

        string += ["</Job>"]

        return "".join(string)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
:mod: DataUsageHandler

.. module: DataUsageHandler

:synopsis: Implementation of the Data Usage service in the DISET framework.
"""
# # from DIRAC
from DIRAC import S_OK
from DIRAC.Core.DISET.RequestHandler import RequestHandler

# # from LHCbDIRAC
from LHCbDIRAC.DataManagementSystem.DB.StorageUsageDB import StorageUsageDB

# # RCSID
# global instance of the StorageUsageDB class
gStorageUsageDB = False


def initializeDataUsageHandler(_serviceInfo):
    """service initalisation."""
    global gStorageUsageDB
    gStorageUsageDB = StorageUsageDB()
    return S_OK()


class DataUsageHandler(RequestHandler):
    """
    .. class:: DataUsageHandler
    """

    types_sendDataUsageReport = [(str,), dict]

    @staticmethod
    def export_sendDataUsageReport(site, directoryDict, status="New"):
        """export of sendDataUsageReport."""
        return gStorageUsageDB.sendDataUsageReport(site, directoryDict, status)

    types_getDataUsageSummary = [(str,), (str,), (str,)]

    @staticmethod
    def export_getDataUsageSummary(startTime, endTime, status):
        """export of getDataUsageSummary."""
        return gStorageUsageDB.getDataUsageSummary(startTime, endTime, status)

    types_getDataUsageForDirectory = [(str,)]

    @staticmethod
    def export_getDataUsageForDirectory(path):
        """export of getDataUsageForDirectory."""
        return gStorageUsageDB.getDataUsageForDirectory(path)

    types_sendDataUsageReport_2 = [(dict)]

    @staticmethod
    def export_sendDataUsageReport_2(directoryDict):
        """export of sendDataUsageReport (new version)"""
        return gStorageUsageDB.sendDataUsageReport_2(directoryDict)

    types_updatePopEntryStatus = [list, (str,)]

    @staticmethod
    def export_updatePopEntryStatus(idList, newStatus):
        """export of updatePopEntryStatus."""
        return gStorageUsageDB.updatePopEntryStatus(idList, newStatus)

    types_insertToDirMetadata = [dict]

    @staticmethod
    def export_insertToDirMetadata(directoryDict):
        """export of insertToDirMetadata."""
        return gStorageUsageDB.insertToDirMetadata(directoryDict)

    types_getDirMetadata = [list]

    @staticmethod
    def export_getDirMetadata(directoryList):
        """export of getDirMetadata."""
        return gStorageUsageDB.getDirMetadata(directoryList)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Utility to construct production LFNs from workflow parameters according to
LHCb conventions.

The methods here are mostly from ancient history and need to be
reviewed, these methods were grouped together as they form the
"interface" for production clients and workflow modules to create LFNs.
"""
import re
import os
import datetime
import copy

from DIRAC import S_OK, S_ERROR, gLogger

gLogger = gLogger.getSubLogger("ProductionData")

#############################################################################


def constructProductionLFNs(paramDict, bkClient=None, quick=True):
    """Used for local testing of a workflow, a temporary measure until LFN
    construction is tidied.

    This works using the workflow commons for on the fly construction.
    """
    try:
        keys = ["PRODUCTION_ID", "JOB_ID", "configVersion", "outputList", "configName", "outputDataFileMask"]
        for k in keys:
            if k not in paramDict:
                return S_ERROR(f"{k} not defined")

        productionID = paramDict["PRODUCTION_ID"]
        jobID = paramDict["JOB_ID"]
        wfConfigName = paramDict["configName"]
        wfConfigVersion = paramDict["configVersion"]
        wfMask = paramDict["outputDataFileMask"]
        if not isinstance(wfMask, list):
            wfMask = [i.lower().strip() for i in wfMask.split(";")]
        outputList = paramDict["outputList"]

        fileTupleList = []
        gLogger.verbose(f"wfConfigName = {wfConfigName}, wfConfigVersion = {wfConfigVersion}, wfMask = {wfMask}")
        for info in outputList:
            try:
                fileName = info["outputDataName"]
            except KeyError:
                # this happens when the parameters are set at runtime (e.g. parametric jobs)
                # should only apply for preSubmission LFNs
                prodID = str(productionID).zfill(8)
                jobID = str(jobID).zfill(8)
                stepInstanceNumber = "1"  # can't be more precise at this stage
                fileName = f"{prodID}_{jobID}_{stepInstanceNumber}"
            fileTupleList.append((fileName, info["outputDataType"]))

        # Strip output data according to file mask
        fileTupleListMasked = _applyMask(wfMask, fileTupleList)
        lfnRoot = _getLFNRoot("", wfConfigName, wfConfigVersion, bkClient, quick=quick)
        gLogger.debug(f"LFN_ROOT is: {lfnRoot}")
        debugRoot = _getLFNRoot("", "debug", wfConfigVersion, bkClient, quick=quick)

        gLogger.verbose(f"LFN_ROOT is: {lfnRoot}")
        if not lfnRoot:
            return S_ERROR("LFN root could not be constructed")

        # Get all LFN(s) to both output data and BK lists at this point (fine for BK)
        outputData = []
        bkLFNs = []
        debugLFNs = []

        # outputData is masked
        for fileTuple in fileTupleListMasked:
            lfn = _makeProductionLFN(str(jobID).zfill(8), lfnRoot, fileTuple, str(productionID).zfill(8))
            outputData.append(lfn)

        # BKLFNs and debugLFNs are not masked
        for fileTuple in fileTupleList:
            lfn = _makeProductionLFN(str(jobID).zfill(8), lfnRoot, fileTuple, str(productionID).zfill(8))
            bkLFNs.append(lfn)
            if debugRoot:
                debugLFNs.append(
                    _makeProductionLFN(str(jobID).zfill(8), debugRoot, fileTuple, str(productionID).zfill(8))
                )

        if debugRoot:
            debugLFNs.append(
                _makeProductionLFN(
                    str(jobID).zfill(8),
                    debugRoot,
                    (f"{str(jobID).zfill(8)}_core", "core"),
                    str(productionID).zfill(8),
                )
            )

        # Get log file path - unique for all modules
        logPath = _makeProductionPath(str(jobID).zfill(8), lfnRoot, "LOG", str(productionID).zfill(8), log=True)
        logFilePath = [logPath]
        logTargetPath = [f"{logPath}/{str(jobID).zfill(8)}.zip"]
        # [ aside, why does makeProductionPath not append the jobID itself ????
        # this is really only used in one place since the logTargetPath is just
        # written to a text file (should be reviewed)... ]

        if not outputData:
            gLogger.info("No output data LFN(s) constructed")
        else:
            gLogger.verbose("Created the following output data LFN(s):\n%s" % ("\n".join(outputData)))
        gLogger.verbose(f"Log file path is:\n{logFilePath[0]}")
        gLogger.verbose(f"Log target path is:\n{logTargetPath[0]}")
        if bkLFNs:
            gLogger.verbose("BookkeepingLFN(s) are:\n%s" % ("\n".join(bkLFNs)))
        if debugLFNs:
            gLogger.verbose("DebugLFN(s) are:\n%s" % ("\n".join(debugLFNs)))
        jobOutputs = {
            "ProductionOutputData": outputData,
            "LogFilePath": logFilePath,
            "LogTargetPath": logTargetPath,
            "BookkeepingLFNs": bkLFNs,
            "DebugLFNs": debugLFNs,
        }
        return S_OK(jobOutputs)

    except Exception as e:
        gLogger.exception(str(e))
        return S_ERROR(e)


#############################################################################


def _applyMask(mask, dataTuplesList):
    """apply the MASK to the dataset."""

    maskedData = copy.deepcopy(dataTuplesList)

    if not isinstance(mask, list):
        mask = [mask]

    if mask != [""]:
        maskLower = [x.lower() for x in mask]

        for dt in dataTuplesList:
            if dt[1].lower() not in maskLower:
                maskedData.remove(dt)

    return maskedData


#############################################################################


def getLogPath(paramDict, bkClient=None, quick=True):
    """Can construct log file paths even if job fails e.g. no output files
    available."""
    try:
        keys = ["PRODUCTION_ID", "JOB_ID", "configName", "configVersion"]
        for k in keys:
            if k not in paramDict:
                return S_ERROR(f"{k} not defined")

        productionID = paramDict["PRODUCTION_ID"]
        jobID = paramDict["JOB_ID"]
        wfConfigName = paramDict["configName"]
        wfConfigVersion = paramDict["configVersion"]

        gLogger.verbose(f"wfConfigName = {wfConfigName}, wfConfigVersion = {wfConfigVersion}")
        lfnRoot = _getLFNRoot("", wfConfigName, wfConfigVersion, bkClient, quick=quick)

        # Get log file path - unique for all modules
        logPath = _makeProductionPath(str(jobID).zfill(8), lfnRoot, "LOG", str(productionID).zfill(8), log=True)
        logFilePath = [logPath]
        logTargetPath = [f"{logPath}/{str(jobID).zfill(8)}.zip"]

        gLogger.verbose(f"Log file path is:\n{logFilePath}")
        gLogger.verbose(f"Log target path is:\n{logTargetPath}")
        jobOutputs = {"LogFilePath": logFilePath, "LogTargetPath": logTargetPath}
        return S_OK(jobOutputs)
    except Exception as e:
        return S_ERROR(e)


#############################################################################


def constructUserLFNs(jobID, owner, outputFiles, outputPath="", prependString=""):
    """This method is used to supplant the standard job wrapper output data
    policy for LHCb.

    The initial convention adopted for user output files is the following:
    /lhcb/user/<initial e.g. f>/<owner e.g. fstagni>/<outputPath>/year_month/<jobID/1000>/<jobID>/fileName

    But, if prependString is set, then the following is adopted:
    /lhcb/user/<initial e.g. f>/<owner e.g. fstagni>/<outputPath>/<jobID>_<prependString>_fileName
    """

    initial = owner[:1]
    timeTup = datetime.date.today().timetuple()
    yearMonth = f"{timeTup[0]}_{str(timeTup[1]).zfill(2)}"
    outputLFNs = {}

    outputPathStr = ""

    # Strip out any leading or trailing slashes but allow fine structure
    if outputPath:
        outputPathList = str(outputPath).split(os.sep)
        newPath = []
        for i in outputPathList:
            if i:
                newPath.append(i)
        outputPathStr = (os.sep).join(newPath)

    if not isinstance(outputFiles, list):
        outputFiles = [outputFiles]

    if prependString:
        for outputFile in outputFiles:
            outputFile = outputFile.replace("LFN:", "")
            lfn = (
                os.sep
                + os.path.join("lhcb", "user", initial, owner, outputPathStr, yearMonth)
                + os.sep
                + str(jobID)
                + "_"
                + prependString
                + "_"
                + os.path.basename(outputFile)
            )
            outputLFNs[outputFile] = lfn
    else:
        for outputFile in outputFiles:
            outputFile = outputFile.replace("LFN:", "")
            lfn = (
                os.sep
                + os.path.join(
                    "lhcb", "user", initial, owner, outputPath, yearMonth, str(int(jobID / 1000)), str(jobID)
                )
                + os.sep
                + os.path.basename(outputFile)
            )
            outputLFNs[outputFile] = lfn

    outputData = list(outputLFNs.values())
    if outputData:
        gLogger.info("Created the following output data LFN(s):\n%s" % ("\n".join(outputData)))
    else:
        gLogger.info("No output LFN(s) constructed")

    return outputData


#############################################################################


def preSubmissionLFNs(jobCommons, jobCode, productionID="1", jobID="2"):
    """Constructs LFNs to be added to the job description prior to submission or
    simply for visual inspection.

    This is a wrapper around constructProductionLFNs used by the
    production clients.
    """
    try:
        outputList = []
        for line in jobCode.split("\n"):
            if line.count("listoutput"):
                outputList += eval(line.split("#")[0].split("=")[-1])

        jobCommons["outputList"] = outputList
        jobCommons["PRODUCTION_ID"] = productionID
        jobCommons["JOB_ID"] = jobID

        gLogger.debug(jobCommons)
        result = constructProductionLFNs(jobCommons)
        if not result["OK"]:
            gLogger.error(result)
        return result
    except Exception as e:
        return S_ERROR(e)


#############################################################################


def _makeProductionPath(jobID, lfnROOT, typeName, prodstring, log=False):
    """Constructs the path in the logical name space where the output data for
    the given production will go."""
    result = lfnROOT + "/" + typeName.upper() + "/" + prodstring + "/"
    if log:
        try:
            jobid = int(jobID)
            jobindex = str(int(jobid / 10000)).zfill(4)
        except Exception:
            jobindex = "0000"
        result += jobindex

    return result


#############################################################################


def _makeProductionLFN(jobID, lfnROOT, filetuple, prodstring):
    """Constructs the logical file name according to LHCb conventions.

    Returns the lfn without 'lfn:' prepended.
    """
    gLogger.debug(
        "Making production LFN for JOB_ID %s, LFN_ROOT %s, prodstring %s for %s"
        % (jobID, lfnROOT, prodstring, str(filetuple))
    )
    try:
        jobid = int(jobID)
        jobindex = str(int(jobid / 10000)).zfill(4)
    except Exception:
        jobindex = "0000"

    fname = filetuple[0]
    if re.search("lfn:", fname) or re.search("LFN:", fname):
        return fname.replace("lfn:", "").replace("LFN:", "")

    return lfnROOT + "/" + filetuple[1].upper() + "/" + prodstring + "/" + jobindex + "/" + fname


#############################################################################


def _getLFNRoot(lfn, namespace="", configVersion=0, bkClient=None, quick=False):
    """return the root path of a given lfn.

    eg : /lhcb/data/CCRC08/00009909 = getLFNRoot(/lhcb/data/CCRC08/00009909/DST/0000/00009909_00003456_2.dst)
    eg : /lhcb/MC/<year>/  = getLFNRoot(None)
    """
    LFN_ROOT = ""

    if not lfn:
        LFN_ROOT = f"/lhcb/{namespace}/{configVersion}"
        gLogger.debug(f"LFN_ROOT will be {LFN_ROOT}")
        return LFN_ROOT

    lfn = [fname.replace(" ", "").replace("LFN:", "") for fname in lfn.split(";")]
    lfnroot = [part for part in lfn[0].split("/") if part]

    if quick:
        for part in lfnroot[0:4]:
            LFN_ROOT += f"/{part}"

    else:
        if not bkClient:
            from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

            bkClient = BookkeepingClient()

        dataTypes = bkClient.getFileTypes({})
        if not dataTypes["OK"]:
            raise Exception(dataTypes["Message"])
        dataTypes = [x[0] for x in dataTypes["Value"]["Records"]]
        gLogger.verbose(f"DataTypes retrieved from the BKK are:\n{', '.join(dataTypes)}")
        gLogger.verbose(f"wf lfn: {lfn}, namespace: {namespace}, configVersion: {configVersion}")

        for part in lfnroot:
            if part not in dataTypes:
                LFN_ROOT += f"/{part}"
            else:
                break

    if re.search("//", LFN_ROOT):
        LFN_ROOT = LFN_ROOT.replace("//", "/")

    if namespace.lower() in ("test", "debug"):
        tmpLfnRoot = LFN_ROOT.split(os.path.sep)
        if len(tmpLfnRoot) > 2:
            tmpLfnRoot[2] = namespace
        else:
            tmpLfnRoot[-1] = namespace

        LFN_ROOT = (os.path.sep).join(
            tmpLfnRoot,
        )

    return LFN_ROOT



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import argparse
import asyncio

import DIRAC
from DIRAC.Core.Security.Properties import PRODUCTION_MANAGEMENT
from rich.console import Console
from rich.prompt import Prompt


def main():
    parser = argparse.ArgumentParser(description="Manage LHCbDIRAC production requests")
    subparsers = parser.add_subparsers(dest="command", required=True)

    subparser = subparsers.add_parser("show-requests", help="Show the current status of the production requests")
    subparser.set_defaults(func=show_requests)
    subparser.add_argument("--execute-actions", action="store_true", help="Execute the actions")

    subparser = subparsers.add_parser("update-metadata", help="Update the metadata stored on GitLab")
    subparser.set_defaults(func=update_metadata)

    subparser = subparsers.add_parser("run-checks", help="Run checks on the production requests")
    subparser.set_defaults(func=run_checks)

    args = parser.parse_args()
    args.func(args)


def show_requests(args):
    DIRAC.initialize()
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools import ProdRequestsGitlabRepo
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.monitoring import (
        analyse_active_productions,
        display_table,
        display_actions,
    )

    if args.execute_actions:
        DIRAC.initialize(security_expression=PRODUCTION_MANAGEMENT)
        match Prompt.ask("Update metadata?", choices=["yes", "no"], default="yes"):
            case "yes":
                update_metadata(args)

    repo = ProdRequestsGitlabRepo(with_auth=args.execute_actions)
    last_update, tables_data, actions = analyse_active_productions(repo)
    console = Console()
    console.rule()
    display_table(console, tables_data)
    console.print(f"Last update: {last_update}")
    console.rule()
    display_actions(console, actions, execute=args.execute_actions)


def update_metadata(args):
    DIRAC.initialize(security_expression=PRODUCTION_MANAGEMENT)
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools import ProdRequestsGitlabRepo
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.monitoring import ACTIVE_PRODUCTION_STATES

    repo = ProdRequestsGitlabRepo(with_auth=True)
    repo.poll(do_status_update=True, states=ACTIVE_PRODUCTION_STATES)


def run_checks(args):
    DIRAC.initialize(security_expression=PRODUCTION_MANAGEMENT)
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools import (
        ProdRequestsGitlabRepo,
        OperationsLogbook,
        do_checks,
    )
    from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.monitoring import (
        analyse_active_productions,
        display_table,
    )

    repo = ProdRequestsGitlabRepo(with_auth=True)
    logbook = OperationsLogbook()
    last_update, tables_data, _ = analyse_active_productions(repo, states=["checking"])

    console = Console()
    console.rule()
    display_table(console, tables_data)
    console.print(f"Last update: {last_update}")
    console.rule()

    request_ids = {
        str(x[1]): x[0]
        for v in tables_data.values()
        for vv in v.values()
        for vvv in vv.values()
        for vvvv in vvv.values()
        for x in vvvv
    }

    asyncio.run(do_checks(console, logbook, repo, request_ids))


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""DIRAC.Workflow.Modules package."""



###############################################################################
# (c) Copyright 2022 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from enum import Enum, StrEnum
from typing import Any, Literal, Annotated, Union

from pydantic import (
    ConfigDict,
    StringConstraints,
    BaseModel as _BaseModel,
    field_validator,
    model_validator,
    Field,
    PositiveInt,
    TypeAdapter,
)

from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
from typing import Annotated


FILETYPE_PATTERN = r"[A-Z0-9\.]+"


class ProductionStates(str, Enum):
    NEW = "New"
    ACTIVE = "Active"
    SUBMITTED = "Submitted"
    PPG_OK = "PPG OK"
    TECH_OK = "Tech OK"
    ACCEPTED = "Accepted"
    DONE = "Done"
    COMPLETED = "Completed"
    CANCELLED = "Cancelled"
    REJECTED = "Rejected"


class BaseModel(_BaseModel):
    model_config = ConfigDict(extra="forbid")


class ProductionStep(BaseModel):
    id: PositiveInt | None = None
    name: str
    processing_pass: str

    class GaudirunOptions(BaseModel):
        command: Annotated[list[str], Field(min_items=1)] | None = None
        files: list[str]
        format: str | None = None
        gaudi_extra_options: str | None = None
        processing_pass: str | None = None

    class LbExecOptions(BaseModel):
        entrypoint: str
        extra_options: dict[str, Any]
        extra_args: list[str] = []

    options: list[str] | GaudirunOptions | LbExecOptions  # TODO the list of str is for legacy compatibility
    options_format: str | None = None  # TODO This should be merged into options
    visible: bool
    multicore: bool = False
    ready: bool = True
    obsolete: bool = False

    class Application(BaseModel):
        name: str
        version: str
        nightly: str | None = None
        binary_tag: str | None = None

    application: Application

    class DataPackage(BaseModel):
        name: str
        version: str

    data_pkgs: list[DataPackage] = []

    class DBTags(BaseModel):
        DDDB: str | None = None
        CondDB: str | None = None
        DQTag: str | None = None

    dbtags: DBTags | None = None

    class FileType(BaseModel):
        type: Annotated[str, StringConstraints(pattern=FILETYPE_PATTERN)]
        visible: bool

    class InputFileType(FileType):
        step_idx: int | None = None

    input: list[InputFileType]
    output: Annotated[list[FileType], Field(min_length=1)]

    @field_validator("input", "output", mode="before")
    @classmethod
    def filetype_default(cls, filetypes, values):  # pylint: disable=no-self-argument
        """Expand a list of file types into the explicit form.

        This is to allow for a convenient list of input types to be specified.
        For example::

          input:
            - type: ["STREAM1.DST", "STREAM2.MDST"]
              visible: true
            - type: STREAM3.DST
              visible: false

        becomes::

          input:
            - type: STREAM1.DST
              visible: true
            - type: STREAM2.MDST
              visible: true
            - type: STREAM3.MDST
              visible: false
        """
        cleaned_filetypes = []
        for filetype in filetypes:
            if isinstance(filetype["type"], list):
                cleaned_filetypes.extend({"type": t, "visible": filetype["visible"]} for t in filetype["type"])
            else:
                cleaned_filetypes.append(filetype)
        return cleaned_filetypes

    @field_validator("input", mode="after")
    @classmethod
    def set_default_input_step_idx(cls, input):
        for idx, input_file in enumerate(input):
            if input_file.step_idx is None:
                input_file.step_idx = idx
        return input

    @field_validator("application", mode="before")
    @classmethod
    def application_default(cls, application, values):  # pylint: disable=no-self-argument
        if isinstance(application, str):
            name, version = application.rsplit("/", 2)
            application = {"name": name, "version": version}
        return application

    @field_validator("data_pkgs", mode="before")
    @classmethod
    def data_pkgs_default(cls, data_pkgs, values):  # pylint: disable=no-self-argument
        cleaned_data_pkgs = []
        for data_pkg in data_pkgs:
            if isinstance(data_pkg, str):
                name, version = data_pkg.rsplit(".", 2)
                cleaned_data_pkgs.append({"name": name, "version": version})
            else:
                cleaned_data_pkgs.append(data_pkg)
        return cleaned_data_pkgs


class InvalidStep(Exception):
    pass


class TransformationDescription(BaseModel):
    type: str
    steps: list[int]
    output_se: str
    input_plugin: str
    group_size: int
    input_data_policy: Literal["protocol"] | Literal["download"]
    output_mode: Literal["Local"] | Literal["Run"] | Literal["Any"]

    output_file_mask: str = ""
    multicore: bool = True
    events: int = -1
    ancestor_depth: int = 0
    cpu: str = "1000000"
    priority: int = 2
    remove_inputs_flags: bool = False
    processors: tuple[int, int] = (0, 0)


class ProductionBase(BaseModel):
    type: str
    id: PositiveInt | None = None
    author: str = Field(default_factory=lambda: returnValueOrRaise(getProxyInfo())["username"])
    priority: Annotated[str, Field(pattern=r"^[12][ab]$")]
    name: Annotated[str, StringConstraints(strip_whitespace=True, min_length=3, max_length=128)]
    inform: list[Annotated[str, StringConstraints(strip_whitespace=True, min_length=3, max_length=50)]] = []
    # TODO: I'm inclined to hardcode the list of known working groups
    wg: str
    comment: str = ""
    state: ProductionStates = ProductionStates.NEW

    steps: Annotated[list[ProductionStep], Field(min_length=1)]

    class SubmissionInfo(BaseModel):
        transforms: list[TransformationDescription]

    # TODO: will eventually be mandatory
    submission_info: SubmissionInfo | None = None

    @model_validator(mode="after")
    def validate_steps(cls, values):  # pylint: disable=no-self-argument
        current_leaves = {ft.type for ft in values.steps[0].output}
        for step in values.steps[1:]:
            for input_filetype in step.input:
                if input_filetype.type not in current_leaves:
                    raise InvalidStep(f"Failed to find input {input_filetype.type!r} for {step!r}")
                current_leaves.remove(input_filetype.type)

            for output_filetype in step.output:
                if output_filetype.type in current_leaves:
                    raise InvalidStep(
                        f"Found producer for {output_filetype.type!r} despite previous instance having not been consumed"
                    )
                current_leaves.add(output_filetype.type)
        return values


class SimulationProduction(ProductionBase):
    type: Literal["Simulation"]
    mc_config_version: str
    sim_condition: str
    fast_simulation_type: Annotated[str, Field(strip_whitespace=True, min_length=4, max_length=32)] = "None"
    # TODO This should move to EventType
    retention_rate: Annotated[float, Field(gt=0, le=1)] = 1
    override_processing_pass: str | None = None

    class EventType(BaseModel):
        id: Annotated[str, Field(pattern=r"[0-9]{8}")]
        num_events: PositiveInt
        num_test_events: PositiveInt = 10

    event_types: Annotated[list[EventType], Field(min_length=1)]


class DataProduction(ProductionBase):
    type: Literal["Sprucing", "AnalysisProduction", "Stripping", "Reconstruction"]

    class InputDataset(BaseModel):
        class BookkeepingQuery(BaseModel):
            configName: str
            configVersion: str
            inFileType: str
            inProPass: str
            # TODO: Accept None and lists on these
            inDataQualityFlag: str = "OK"
            inExtendedDQOK: Annotated[list[str], Field(min_items=1)] | None = None
            inProductionID: str = "ALL"
            inTCKs: str = "ALL"
            inSMOG2State: Annotated[list[str], Field(min_items=1)] | None = None

        conditions_dict: BookkeepingQuery
        conditions_description: str
        event_type: Annotated[str, StringConstraints(pattern=r"[0-9]{8}")]

        class LaunchParameters(BaseModel):
            """Extra info which is currently needed when making transformations from a request"""

            run_numbers: list[str] | None = None
            sample_max_md5: Annotated[str, StringConstraints(pattern=r"[A-Z0-9]{32}")] | None = None
            sample_seed_md5: Annotated[str, StringConstraints(pattern=r"[A-Z0-9]{32}")] | None = None
            start_run: int | None = None
            end_run: int | None = None

        launch_parameters: Annotated[LaunchParameters, Field(default_factory=LaunchParameters)]

    input_dataset: InputDataset


Production = Annotated[Union[SimulationProduction, DataProduction], Field(discriminator="type")]


def parse_obj(obj: Any) -> Production:
    adapter = TypeAdapter(Production)
    return adapter.validate_python(obj)



#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Get statistical information on a dataset."""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerBKSwitches()
    dmScript.registerFileSwitches()

    Script.registerSwitch("", "TriggerRate", "   For RAW files, returns the trigger rate")
    Script.registerSwitch("", "ListRuns", "   Give a list of runs (to be used with --Trigger)")
    Script.registerSwitch("", "ListFills", "   Give a list of fills (to be used with --Trigger)")
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile]",
            ]
        )
    )

    Script.parseCommandLine(ignoreErrors=False)

    from LHCbDIRAC.BookkeepingSystem.Client.ScriptExecutors import executeGetStats

    executeGetStats(dmScript)


if __name__ == "__main__":
    main()
